/*
  matrix2x2.c 
  Kompilation:
    gcc -o matrix2x2 matrix2x2.c \
       -I. -O2 -lm
  Ausfhrung:
  matrix <type> <size> <loops>
*/

#include <3dnow.h>
#include <math.h>

int main(int argc, char* argv[]) {

  int type  = atoi(argv[1]);
  int size  = atoi(argv[2]);
  int loops = atoi(argv[3]);
  int i,j;

  mmx_t A[2],b,x[size],y[size];

  /* phi in RAD */
  float phi = 45 * 3.1515926/180;

  /* erster Eingangsvektor */
  x[0].s[0] = 4;
  x[0].s[1] = 3;

  /* Rotationsmatrix A */
  A[0].s[0] =  cos(phi); /* a11 */
  A[0].s[1] = -sin(phi); /* a12 */
  A[1].s[0] =  sin(phi); /* a21 */
  A[1].s[1] =  cos(phi); /* a22 */

  /* Translationsvektor b */
  b.s[0] = -2; /* b1 */
  b.s[1] =  2; /* b2 */

  /* ----- Traditionell ----- */
  if(type==1) {
    for(j=0;j<loops;j++) {
      for(i=0;i<size;i++) {

	y[i].s[0]=A[0].s[0]*x[i].s[0]
	  +A[0].s[1]*x[i].s[1]+b.s[0];
	y[i].s[1]=A[1].s[0]*x[i].s[0]
	  +A[1].s[1]*x[i].s[1]+b.s[1];
	
      };
    };
  };

  /* 3DNow!-Support prfen */
  if(mm_support()!=5) exit(1);

  /* ----- mit 3DNow!-Makros ----- */
  if(type==2) {
    for(j=0;j<loops;j++) {
      for(i=0;i<size;i++) {
	
	mmx_t t1=x[i],t2=x[i];
	pfmul(A[0],t1);
	pfmul(A[1],t2);
	pfacc(t2,t1);
	pfadd(b,t1);
	y[i]=t1;
	
      };
    };
  };
  

  /* ----- mit Register-Makros ----- */
  if(type==3) {
    for(j=0;j<loops;j++) {

      movq_m2r(A[0],mm0);
      movq_m2r(A[1],mm1);
      movq_m2r(b,mm2);

      for(i=0;i<size;i++) {

	movq_m2r(x[i],mm3);
	movq_r2r(mm3,mm4);
	mmx_r2r(pfmul,mm0,mm3);
	mmx_r2r(pfmul,mm1,mm4);
	mmx_r2r(pfacc,mm4,mm3);
	mmx_r2r(pfadd,mm2,mm3);
	movq_r2m(mm3,y[i]);
	
      };
    };
  };
  
  /* Pipline-Sync */
  femms();

  /* ersten Ausgangsvektor ausgeben */
  printf("%f %f\n",y[0].s[0],y[0].s[1]);

  return 0;
};




