#include <3dnow.h>

#define SIZE 1
#define LOOPS 100000000


typedef struct {
  float a11,a12,a13,a14;
  float a21,a22,a23,a24;
  float a31,a32,a33,a34;
  float a41,a42,a43,a44;
} matrix;

typedef struct {
  float x1,x2,x3,x4;
} vektor;

inline void mult(matrix *m,vektor *x,vektor *y) {
  y->x1=m->a11*x->x1+m->a12*x->x2+m->a13*x->x3+m->a14*x->x4;
  y->x2=m->a21*x->x1+m->a22*x->x2+m->a23*x->x3+m->a24*x->x4;
  y->x3=m->a31*x->x1+m->a32*x->x2+m->a33*x->x3+m->a34*x->x4;
  y->x4=m->a41*x->x1+m->a42*x->x2+m->a43*x->x3+m->a44*x->x4;
};



typedef struct {
  mmx_t z11,z12;
  mmx_t z21,z22;
  mmx_t z31,z32;
  mmx_t z41,z42;
} mm_matrix;


typedef struct {
  mmx_t s1,s2;
} mm_vektor;

inline void mm_mult(mm_matrix *m,mm_vektor *x,mm_vektor *y) {

  movq_m2r(x->s1,mm6);    /* mm6 = s1      */
  movq_m2r(x->s2,mm7);    /* mm7 = s2      */

  movq_m2r(m->z11,mm0);   /* mm0 = z11     */
  mmx_r2r(pfmul,mm6,mm0); /* s1*z11 -> mm0 */
  movq_m2r(m->z12,mm1);   /* mm1 = z12     */
  mmx_r2r(pfmul,mm7,mm1); /* s2*z12 -> mm1 */
  mmx_r2r(pfadd,mm1,mm0); /* s1*z11+s2*z12 -> mm0 */;

  movq_m2r(m->z21,mm1);   /* mm1 = z21     */
  mmx_r2r(pfmul,mm6,mm1); /* s1*z21 -> mm1 */
  movq_m2r(m->z22,mm2);   /* mm2 = z22     */
  mmx_r2r(pfmul,mm7,mm2); /* s2*z22 -> mm2 */
  mmx_r2r(pfadd,mm2,mm1); /* s1*z21+s2*z22 -> mm1 */

  mmx_r2r(pfacc,mm1,mm0); /* y->s1 */
  movq_r2m(mm0,y->s1);

  movq_m2r(m->z31,mm0);   /* mm0 = z31     */
  mmx_r2r(pfmul,mm6,mm0); /* s1*z31 -> mm0 */
  movq_m2r(m->z32,mm1);   /* mm1 = z32     */
  mmx_r2r(pfmul,mm7,mm1); /* s2*z32 -> mm1 */
  mmx_r2r(pfadd,mm1,mm0); /* s1*z31+s2*z32 -> mm0 */

  movq_m2r(m->z41,mm1);   /* mm1 = z41     */
  mmx_r2r(pfmul,mm6,mm1); /* s1*z41 -> mm1 */
  movq_m2r(m->z42,mm2);   /* mm2 = z42     */
  mmx_r2r(pfmul,mm7,mm2); /* s2*z42 -> mm2 */
  mmx_r2r(pfadd,mm2,mm1); /* s1*z41+s2*z42 -> mm1 */

  mmx_r2r(pfacc,mm1,mm0); /* y->s2 */
  movq_r2m(mm0,y->s2);

};


inline void mm_mult2(mm_matrix *m,mm_vektor *x,mm_vektor *y) {

  movq_m2r(x->s1,mm6);    /* mm6 = s1      */
  movq_m2r(x->s2,mm7);    /* mm7 = s2      */

  movq_m2r(m->z11,mm0);   /* mm0 = z11     */
  movq_m2r(m->z12,mm1);   /* mm1 = z12     */
  mmx_r2r(pfmul,mm6,mm0); /* s1*z11 -> mm0 */
  mmx_r2r(pfmul,mm7,mm1); /* s2*z12 -> mm1 */
  mmx_r2r(pfadd,mm1,mm0); /* s1*z11+s2*z12 -> mm0 */;

  movq_m2r(m->z21,mm1);   /* mm1 = z21     */
  movq_m2r(m->z22,mm2);   /* mm2 = z22     */
  mmx_r2r(pfmul,mm6,mm1); /* s1*z21 -> mm1 */
  mmx_r2r(pfmul,mm7,mm2); /* s2*z22 -> mm2 */
  mmx_r2r(pfadd,mm2,mm1); /* s1*z21+s2*z22 -> mm1 */

  mmx_r2r(pfacc,mm1,mm0); /* y->s1 */
  movq_r2m(mm0,y->s1);

  movq_m2r(m->z31,mm0);   /* mm0 = z31     */
  mmx_r2r(pfmul,mm6,mm0); /* s1*z31 -> mm0 */
  movq_m2r(m->z32,mm1);   /* mm1 = z32     */
  mmx_r2r(pfmul,mm7,mm1); /* s2*z32 -> mm1 */
  mmx_r2r(pfadd,mm1,mm0); /* s1*z31+s2*z32 -> mm0 */

  movq_m2r(m->z41,mm1);   /* mm1 = z41     */
  mmx_r2r(pfmul,mm6,mm1); /* s1*z41 -> mm1 */
  movq_m2r(m->z42,mm2);   /* mm2 = z42     */
  mmx_r2r(pfmul,mm7,mm2); /* s2*z42 -> mm2 */
  mmx_r2r(pfadd,mm2,mm1); /* s1*z41+s2*z42 -> mm1 */

  mmx_r2r(pfacc,mm1,mm0); /* y->s2 */
  movq_r2m(mm0,y->s2);

};

int main(void) {
  int i,j;

  matrix m[SIZE] = {1,2,4,2,
	      3,1,2,1,
	      3,3,1,5,
	      5,6,7,1};

  vektor x[SIZE] = {6,7,2,4},y[SIZE];

  for(j=0;j<LOOPS;j++) {
    for(i=0;i<SIZE;i++) {


      mm_mult2((mm_matrix*)&m[0],(mm_vektor*)&x[0],(mm_vektor*)&y[0]);
/*
      mult(&m[0],&x[0],&y[0]);
*/
    };
  };

  femms();
  /*
  printf("%f %f %f %f\n",y.x1,y.x2,y.x3,y.x4);
  */

  return 0;
};







